from pydantic import BaseModel, Field, field_validator
from typing import List, Optional, Dict, Any, Union

class ResultInfo(BaseModel):
    status: int = Field(..., alias="STATUS")
    error_msg: str = Field(..., alias="ERROR_MSG")
    date: str = Field(..., alias="DATE")

class Parameter(BaseModel):
    lang: str = Field(..., alias="LANG")
    stats_data_id: str = Field(..., alias="STATS_DATA_ID")
    data_format: str = Field(..., alias="DATA_FORMAT")
    start_position: int = Field(..., alias="START_POSITION")
    limit: Optional[int] = Field(None, alias="LIMIT")
    metaget_flg: str = Field(..., alias="METAGET_FLG")
    cnt_get_flg: Optional[str] = Field(None, alias="CNT_GET_FLG")

class StatName(BaseModel):
    code: str = Field(..., alias="@code")
    value: str = Field(..., alias="$")

class GovOrg(BaseModel):
    code: str = Field(..., alias="@code")
    value: str = Field(..., alias="$")

class Title(BaseModel):
    no: str = Field(..., alias="@no")
    value: str = Field(..., alias="$")

class Category(BaseModel):
    code: str = Field(..., alias="@code")
    value: str = Field(..., alias="$")

class TableInfo(BaseModel):
    id: str = Field(..., alias="@id")
    stat_name: StatName = Field(..., alias="STAT_NAME")
    gov_org: GovOrg = Field(..., alias="GOV_ORG")
    statistics_name: str = Field(..., alias="STATISTICS_NAME")
    title: Title = Field(..., alias="TITLE")
    cycle: str = Field(..., alias="CYCLE")
    survey_date: Union[str, int] = Field(..., alias="SURVEY_DATE")
    open_date: str = Field(..., alias="OPEN_DATE")
    small_area: Union[str, int] = Field(..., alias="SMALL_AREA")
    collect_area: str = Field(..., alias="COLLECT_AREA")
    main_category: Category = Field(..., alias="MAIN_CATEGORY")
    sub_category: Category = Field(..., alias="SUB_CATEGORY")
    overall_total_number: int = Field(..., alias="OVERALL_TOTAL_NUMBER")
    updated_date: str = Field(..., alias="UPDATED_DATE")

class Class(BaseModel):
    code: str = Field(..., alias="@code")
    name: str = Field(..., alias="@name")
    level: str = Field(..., alias="@level")
    unit: Optional[str] = Field(None, alias="@unit")
    parent_code: Optional[str] = Field(None, alias="@parentCode")

class ClassObj(BaseModel):
    id: str = Field(..., alias="@id")
    name: str = Field(..., alias="@name")
    class_: Union[List[Class], Class] = Field(..., alias="CLASS")

class ClassInfo(BaseModel):
    class_obj: List[ClassObj] = Field(..., alias="CLASS_OBJ")

class Note(BaseModel):
    char: str = Field(..., alias="@char")
    value: str = Field(..., alias="$")

class Annotation(BaseModel):
    annotation: str = Field(..., alias="@annotation")
    value: str = Field(..., alias="$")

class Value(BaseModel):
    tab: Optional[str] = Field(None, alias="@tab")
    cat01: Optional[str] = Field(None, alias="@cat01")
    cat02: Optional[str] = Field(None, alias="@cat02")
    cat03: Optional[str] = Field(None, alias="@cat03")
    cat04: Optional[str] = Field(None, alias="@cat04")
    cat05: Optional[str] = Field(None, alias="@cat05")
    cat06: Optional[str] = Field(None, alias="@cat06")
    cat07: Optional[str] = Field(None, alias="@cat07")
    cat08: Optional[str] = Field(None, alias="@cat08")
    cat09: Optional[str] = Field(None, alias="@cat09")
    cat10: Optional[str] = Field(None, alias="@cat10")
    cat11: Optional[str] = Field(None, alias="@cat11")
    cat12: Optional[str] = Field(None, alias="@cat12")
    cat13: Optional[str] = Field(None, alias="@cat13")
    cat14: Optional[str] = Field(None, alias="@cat14")
    cat15: Optional[str] = Field(None, alias="@cat15")
    area: Optional[str] = Field(None, alias="@area")
    time: str = Field(..., alias="@time")
    unit: str = Field(..., alias="@unit")
    value: str = Field(..., alias="$")


class DataInfo(BaseModel):
    note: List[Note] = Field(..., alias="NOTE")
    annotation: Optional[List[Annotation]] = Field(None, alias="ANNOTATION")
    value: List[Value] = Field(..., alias="VALUE")
    @field_validator("note", mode="before")
    @classmethod
    def ensure_list(cls, v: Union[dict, list]) -> list:
        return [v] if isinstance(v, dict) else v
    
class ResultInf(BaseModel):
    total_number: int = Field(..., alias="TOTAL_NUMBER")
    from_number: int = Field(..., alias="FROM_NUMBER")
    to_number: int = Field(..., alias="TO_NUMBER")
    next_key: Optional[int] = Field(None, alias="NEXT_KEY")

class StatisticalData(BaseModel):
    result_inf: ResultInf = Field(..., alias="RESULT_INF")
    table_inf: TableInfo = Field(..., alias="TABLE_INF")
    class_inf: ClassInfo = Field(..., alias="CLASS_INF")
    data_inf: DataInfo = Field(..., alias="DATA_INF")

class GetStatsData(BaseModel):
    result: ResultInfo = Field(..., alias="RESULT")
    parameter: Parameter = Field(..., alias="PARAMETER")
    statistical_data: StatisticalData = Field(..., alias="STATISTICAL_DATA")

class GetStatsDataResponse(BaseModel):
    get_stats_data: GetStatsData = Field(..., alias="GET_STATS_DATA")



import requests
from pydantic import BaseModel, Field
from typing import List, Optional, Dict, Any
from enum import Enum
import json


class Language(str, Enum):
    JAPANESE = "J"
    ENGLISH = "E"


class DataFormat(str, Enum):
    XML = "X"
    JSON = "J"
    JSONP = "P"
    CSV = "C"


class StatsListParam(BaseModel):
    surveyYears: Optional[str] = None
    openYears: Optional[str] = None
    statsField: Optional[str] = None
    statsCode: Optional[str] = None
    searchWord: Optional[str] = None
    searchKind: Optional[str] = None
    collectArea: Optional[str] = None
    explanationGetFlg: Optional[str] = None
    statsNameList: Optional[str] = None
    startPosition: Optional[int] = None
    limit: Optional[int] = None
    updatedDate: Optional[str] = None


class MetaInfoParam(BaseModel):
    statsDataId: str
    explanationGetFlg: Optional[str] = None


class StatsDataParam(BaseModel):
    dataSetId: Optional[str] = None
    statsDataId: Optional[str] = None
    lvTab: Optional[str] = None
    cdTab: Optional[str] = None
    cdTabFrom: Optional[str] = None
    cdTabTo: Optional[str] = None
    lvTime: Optional[str] = None
    cdTime: Optional[str] = None
    cdTimeFrom: Optional[str] = None
    cdTimeTo: Optional[str] = None
    lvArea: Optional[str] = None
    cdArea: Optional[str] = None
    cdAreaFrom: Optional[str] = None
    cdAreaTo: Optional[str] = None
    lvCat01: Optional[str] = None
    cdCat01: Optional[str] = None
    cdCat01From: Optional[str] = None
    cdCat01To: Optional[str] = None
    startPosition: Optional[int] = None
    limit: Optional[int] = None
    metaGetFlg: Optional[str] = None
    cntGetFlg: Optional[str] = None
    explanationGetFlg: Optional[str] = None
    annotationGetFlg: Optional[str] = None
    replaceSpChar: Optional[str] = None


class EstatsAPIClient:
    BASE_URL = "https://api.e-stat.go.jp/rest/3.0/app/json"

    def __init__(self, app_id: str):
        self.app_id = app_id
        self.session = requests.Session()
        

    def _make_request(self, endpoint: str, params: Dict[str, Any]) -> dict:
        url = f"{self.BASE_URL}/{endpoint}"
        params["appId"] = self.app_id
        response = self.session.get(url, params=params)
        response.raise_for_status()
        res = json.loads(response.content.decode('shift-jis'))
        
        return res

    def get_stats_list(
        self,
        params: StatsListParam,
        lang: Language = Language.JAPANESE,
        data_format: DataFormat = DataFormat.JSON,
    ) -> dict:
        endpoint = "getDataCatalog"
        params_dict = params.model_dump(exclude_none=True)
        params_dict["lang"] = lang.value
        params_dict["dataFormat"] = data_format.value
        return self._make_request(endpoint, params_dict)

    def get_stats_data(
        self,
        params: StatsDataParam,
        lang: Language = Language.JAPANESE,
        data_format: DataFormat = DataFormat.JSON,
    ) -> dict:
        endpoint = "getStatsData"
        params_dict = params.model_dump(exclude_none=True)
        params_dict["lang"] = lang.value
        params_dict["dataFormat"] = data_format.value
        return self._make_request(endpoint, params_dict)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.session.close()


import pytest
import json # 標準のjsonライブラリをインポート
import os # osモジュールをインポート
from dotenv import load_dotenv # dotenvからload_dotenvをインポート
from estats_client.models.client import EstatsAPIClient, StatsDataParam
from estats_client.models.result import GetStatsDataResponse

# .envファイルから環境変数を読み込む
load_dotenv()

# 環境変数からAPP_IDを取得
APP_ID = os.getenv("APP_ID")

# APP_IDが設定されていない場合はエラーとする
if APP_ID is None:
    raise ValueError("環境変数 APP_ID が設定されていません。.envファイルを確認してください。")


def test_get_stats_data_success():
    """
    get_stats_data APIを呼び出し、正常なレスポンスが返ることを確認するテスト。
    """
    client = EstatsAPIClient(app_id=APP_ID)
    params = StatsDataParam(statsDataId="0003173901")

    try:
        response_json_dict = client.get_stats_data(params=params) # APIからのレスポンスは辞書型
        # pydanticモデルに変換してバリデーションも兼ねる
        response_model = GetStatsDataResponse(**response_json_dict)

        # APIの仕様に基づき、RESULT.STATUSが0であることを確認
        assert response_model.get_stats_data.result.status == 0, \
            f"API Error: {response_model.get_stats_data.result.error_msg}"

        # データが取得できていることを確認（例: VALUEのリストが空でないこと）
        assert len(response_model.get_stats_data.statistical_data.data_inf.value) > 0, \
            "No data values returned"

        print("API Response:")
        # model_dump()で辞書に変換し、json.dumpsでensure_ascii=Falseを指定してJSON文字列に変換
        print(json.dumps(response_model.model_dump(), indent=2, ensure_ascii=False))

    except Exception as e:
        pytest.fail(f"API request failed: {e}")


from pydantic import BaseModel, Field, field_validator
from typing import List, Optional, Union
from typing import List, Any
from pydantic import BaseModel, Field


def ensure_list(v: Any) -> List[Any]:
    if isinstance(v, list):
        return v
    return [v]

class ResultInfo(BaseModel):
    status: int = Field(..., alias="STATUS")
    error_msg: str = Field(..., alias="ERROR_MSG")
    date: str = Field(..., alias="DATE")

class Parameter(BaseModel):
    lang: str = Field(..., alias="LANG")
    data_format: str = Field(..., alias="DATA_FORMAT")
    limit: int = Field(..., alias="LIMIT")

class ResultInf(BaseModel):
    from_number: int = Field(..., alias="FROM_NUMBER")
    to_number: int = Field(..., alias="TO_NUMBER")
    next_key: Optional[int] = Field(None, alias="NEXT_KEY")

class StatName(BaseModel):
    code: str = Field(..., alias="@code")
    value: str = Field(..., alias="$")

class Organization(BaseModel):
    code: str = Field(..., alias="@code")
    value: str = Field(..., alias="$")

class Title(BaseModel):
    name: str = Field(..., alias="NAME")
    tabulation_category: str = Field(..., alias="TABULATION_CATEGORY")
    tabulation_sub_category1: str = Field(..., alias="TABULATION_SUB_CATEGORY1")
    tabulation_sub_category2: str = Field(..., alias="TABULATION_SUB_CATEGORY2")
    tabulation_sub_category3: str = Field(..., alias="TABULATION_SUB_CATEGORY3")
    tabulation_sub_category4: str = Field(..., alias="TABULATION_SUB_CATEGORY4")
    tabulation_sub_category5: str = Field(..., alias="TABULATION_SUB_CATEGORY5")
    cycle: str = Field(..., alias="CYCLE")
    survey_date: str|int = Field(..., alias="SURVEY_DATE")
    collect_area: str = Field(..., alias="COLLECT_AREA")

class Dataset(BaseModel):
    stat_name: StatName = Field(..., alias="STAT_NAME")
    organization: Organization = Field(..., alias="ORGANIZATION")
    title: Title = Field(..., alias="TITLE")
    description: str = Field(..., alias="DESCRIPTION")
    publisher: str = Field(..., alias="PUBLISHER")
    contact_point: str = Field(..., alias="CONTACT_POINT")
    creator: str = Field(..., alias="CREATOR")
    release_date: str = Field(..., alias="RELEASE_DATE")
    last_modified_date: str = Field(..., alias="LAST_MODIFIED_DATE")
    frequency_of_update: str = Field(..., alias="FREQUENCY_OF_UPDATE")
    landing_page: str = Field(..., alias="LANDING_PAGE")

class ResourceTitle(BaseModel):
    name: str = Field(..., alias="NAME")
    table_category: Optional[str] = Field(None, alias="TABLE_CATEGORY")
    table_no: Optional[str|int] = Field(None, alias="TABLE_NO")
    table_name: str = Field(..., alias="TABLE_NAME")
    table_explanation: Optional[str] = Field(None, alias="TABLE_EXPLANATION")
    table_sub_category1: Optional[str] = Field(None, alias="TABLE_SUB_CATEGORY1")
    table_sub_category2: Optional[str] = Field(None, alias="TABLE_SUB_CATEGORY2")
    table_sub_category3: Optional[str] = Field(None, alias="TABLE_SUB_CATEGORY3")

class Resource(BaseModel):
    id: str = Field(..., alias="@id")
    title: ResourceTitle = Field(..., alias="TITLE")
    url: str = Field(..., alias="URL")
    description: str = Field(..., alias="DESCRIPTION")
    format: str = Field(..., alias="FORMAT")
    release_date: str = Field(..., alias="RELEASE_DATE")
    last_modified_date: str = Field(..., alias="LAST_MODIFIED_DATE")
    resource_licence_id: str = Field(..., alias="RESOURCE_LICENCE_ID")
    language: str = Field(..., alias="LANGUAGE")




class Resources(BaseModel):
    resource: List[Resource]= Field(alias="RESOURCE",)
    @field_validator("resource", mode="before")
    @classmethod
    def ensure_list(cls, v: Union[dict, list]) -> list:
        return [v] if isinstance(v, dict) else v

class DataCatalogInf(BaseModel):
    id: str = Field(..., alias="@id")
    dataset: Dataset = Field(..., alias="DATASET")
    resources: Resources = Field(..., alias="RESOURCES")

class DataCatalogListInf(BaseModel):
    number: int = Field(..., alias="NUMBER")
    result_inf: ResultInf = Field(..., alias="RESULT_INF")
    data_catalog_inf: List[DataCatalogInf] = Field(..., alias="DATA_CATALOG_INF")

class GetDataCatalog(BaseModel):
    result: ResultInfo = Field(..., alias="RESULT")
    parameter: Parameter = Field(..., alias="PARAMETER")
    data_catalog_list_inf: DataCatalogListInf = Field(..., alias="DATA_CATALOG_LIST_INF")

class GetDataCatalogResponse(BaseModel):
    get_data_catalog: GetDataCatalog = Field(..., alias="GET_DATA_CATALOG")


import os
from pathlib import Path
from estats_client.models.client import EstatsAPIClient, StatsListParam
from estats_client.models.result_get_catalog import DataCatalogInf, GetDataCatalogResponse
import pandas as pd
import json
from tenacity import retry, stop_after_attempt, wait_exponential, retry_if_exception_type

@retry(
    stop=stop_after_attempt(5),
    wait=wait_exponential(multiplier=1, min=4, max=60),
    retry=retry_if_exception_type((ValueError, Exception))
)
def get_stats_list_with_retry(client: EstatsAPIClient, params: StatsListParam):
    return client.get_stats_list(params)

def get_all_catalog(client: EstatsAPIClient, file: Path, start_position=0) -> int | None:
    try:
        j = get_stats_list_with_retry(client, StatsListParam(limit=100, startPosition=start_position))
        if j['GET_DATA_CATALOG']['RESULT']['STATUS'] == 0:
            data = j['GET_DATA_CATALOG']['DATA_CATALOG_LIST_INF']['DATA_CATALOG_INF']

            with file.open("a", encoding="utf-8") as fi:
                for d in data:
                    json.dump(d, fi, ensure_ascii=False)
                    fi.write("\n")
            
            if 'NEXT_KEY' not in j['GET_DATA_CATALOG']['DATA_CATALOG_LIST_INF']['RESULT_INF']:
                return None
            
            next_key = j['GET_DATA_CATALOG']['DATA_CATALOG_LIST_INF']['RESULT_INF']['NEXT_KEY']
            return int(next_key)
        else:
            raise ValueError(j['GET_DATA_CATALOG']['RESULT']['ERROR_MSG'])
    except Exception as e:
        print(f"An error occurred: {str(e)}. Retrying...")
        raise